import datetime
import hashlib
import re
from enum import Enum
from typing import NamedTuple, List, Optional

from pscr_common.api.models import (
    with_swagger_definition,
    with_from_dict,
    ServerProvided,
    ValidationError,
)
from pscr_common.utils import EPOCH

from .helpers import VALID_ID_RE, InvalidReportId, InvalidIOCId, InvalidReportSeverity


@with_swagger_definition
@with_from_dict
class FeedInfo(NamedTuple):
    name: str
    owner: str
    provider_url: str
    summary: str
    category: str
    alertable: Optional[bool] = True
    source_label: Optional[str] = None
    access: ServerProvided[str] = None
    id: ServerProvided[str] = None
    reports_count: Optional[int] = None
    scope: Optional[List[str]] = None


@with_swagger_definition
@with_from_dict
class QueryIOC(NamedTuple):
    index_type: Optional[str]
    search_query: str


@with_swagger_definition
@with_from_dict
class IOCs(NamedTuple):
    md5: Optional[List[str]] = None
    ipv4: Optional[List[str]] = None
    ipv6: Optional[List[str]] = None
    dns: Optional[List[str]] = None
    query: Optional[List[QueryIOC]] = None


class MatchType(Enum):
    EQUALITY = "equality"
    REGEX = "regex"
    QUERY = "query"


@with_swagger_definition
@with_from_dict
class IOC_V2(NamedTuple):
    id: str
    match_type: MatchType
    values: List[str]
    field: Optional[str] = None
    link: Optional[str] = None

    @staticmethod
    def _validate(ioc):
        if not re.match(VALID_ID_RE, ioc.id):
            raise InvalidIOCId(ioc.id)

        valid_match_types = [mt.value for mt in MatchType]

        if not ioc.field and ioc.match_type != "query":
            raise ValidationError("Must provide 'field' for non-query match_type")

        if ioc.match_type not in valid_match_types:
            raise ValidationError(
                f"Unknown match_type: '{ioc.match_type}'.  Expected one of [{', '.join(valid_match_types)}]."
            )


@with_swagger_definition
@with_from_dict
class Report(NamedTuple):
    id: str
    timestamp: datetime.datetime
    title: str
    description: str
    severity: int
    link: Optional[str] = None
    tags: Optional[List[str]] = None
    iocs: Optional[IOCs] = None
    iocs_v2: Optional[List[IOC_V2]] = None
    iocs_total_count: Optional[int] = None
    visibility: Optional[str] = None

    def get_queries(self):
        queries = []
        if self.iocs_v2:
            for ioc in self.iocs_v2:
                if ioc.match_type == MatchType.QUERY.value:
                    queries.extend(ioc.values)

        return queries

    def get_equality_fields(self):
        iocs = []
        if self.iocs_v2:
            for ioc in self.iocs_v2:
                if ioc.match_type == MatchType.EQUALITY.value:
                    iocs.append(ioc)
        return iocs

    @staticmethod
    def _translate(d):
        if "severity" in d:
            d.pop("score", None)
        elif "score" in d:
            score = int(d.pop("score", 0))
            d["severity"] = int(score / 10)

        iocs_v2 = d.pop("iocs_v2", None) or []
        iocs_v2.extend(Report._iocs_to_v2(d.pop("iocs", None)))
        d["iocs_v2"] = iocs_v2

        return d

    @staticmethod
    def _validate(report):
        if not re.match(VALID_ID_RE, report.id):
            raise InvalidReportId(report.id)

        if report.iocs and report.iocs_v2:
            raise ValidationError("Cannot provide both 'iocs' and 'iocs_v2'")

        if report.severity > 10 or report.severity < 1:
            raise InvalidReportSeverity(report.severity)

        if not report.title:
            raise ValidationError("Must provide non-empty 'title' for report")

        if report.tags:
            for tag in report.tags:
                if tag:
                    if not tag.strip():
                        raise ValidationError("Cannot provide empty 'tag' for report")
                else:
                    raise ValidationError("Cannot provide Null 'tag' for report")

        if (report.timestamp - EPOCH).days < 0:
            raise ValidationError("Must provide non-negative 'timestamp' for report")

    @staticmethod
    def _iocs_to_v2(iocs):
        if not iocs:
            return []

        def get_ioc_hash(ioc_type, vals):
            md5 = hashlib.md5(usedforsecurity=False)
            md5.update(ioc_type.encode())
            for s in sorted(vals):
                md5.update(s.encode())
            return md5.hexdigest()

        iocs_v2 = []

        if "ipv4" in iocs:
            iocs_v2.append(
                {
                    "id": get_ioc_hash("ipv4", iocs["ipv4"]),
                    "match_type": "equality",
                    "field": "netconn_ipv4",
                    "values": iocs["ipv4"],
                }
            )
        if "ipv6" in iocs:
            iocs_v2.append(
                {
                    "id": get_ioc_hash("ipv6", iocs["ipv6"]),
                    "match_type": "equality",
                    "field": "netconn_ipv6",
                    "values": iocs["ipv6"],
                }
            )

        if "md5" in iocs:
            iocs_v2.append(
                {
                    "id": get_ioc_hash("md5", iocs["md5"]),
                    "match_type": "equality",
                    "field": "process_hash",
                    "values": iocs["md5"],
                }
            )

        if "dns" in iocs:
            iocs_v2.append(
                {
                    "id": get_ioc_hash("dns", iocs["dns"]),
                    "match_type": "equality",
                    "field": "netconn_domain",
                    "values": iocs["dns"],
                }
            )

        if "query" in iocs:
            values = [q["search_query"] for q in iocs["query"]]
            iocs_v2.append(
                {
                    "id": get_ioc_hash("query", values),
                    "match_type": "query",
                    "values": values,
                }
            )

        return iocs_v2


@with_swagger_definition
@with_from_dict
class Feed(NamedTuple):
    feedinfo: FeedInfo
    reports: List[Report]
