import asyncio
import datetime
import ipaddress

from aio_pyformance.registry import MetricsRegistry
from aiohttp.web_exceptions import HTTPBadRequest
import base64
import logging
import re
import rapidjson as json

from pscr_common.utils import chunker, get_utcnow_with_tzinfo
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.exc import IntegrityError
from sqlalchemy import delete, select, or_, alias, tuple_, exists, column, func
from typing import List, Dict, Set, Iterable, Union, Tuple
import uuid

from sqlalchemy.orm import aliased
from feed_manager.api.models import Report as ApiReport, IOC_V2
from feed_manager.engine.errors import EntityNotFound, NotPermissioned
from feed_manager.engine.org_watchlist_state_updater import OrgWatchlistStateUpdater
from feed_manager.sql.models import (
    Report as DbReport,
    IOC,
    ReportClassifier,
    Classifier,
    ClassifierWatchlist,
    ReportWatchlist,
)
from feed_manager.utils.config import Config
from feed_manager.utils.db import (
    ReadWriteDbPool,
    is_fk_violation_exception,
    is_unique_violation_exception,
)
from feed_manager.utils.search_client import SearchClient

_logger = logging.getLogger(__name__)


class InvalidQuery(HTTPBadRequest):
    def __init__(self, query):
        super(InvalidQuery, self).__init__(
            reason=f"Query '{query}' is invalid or includes a term without a field name"
        )


class InvalidField(HTTPBadRequest):
    def __init__(self, field):
        super(InvalidField, self).__init__(
            reason=f"Field '{field}' is invalid or has an invalid value"
        )


class ReportNotPermissioned(NotPermissioned):
    def __init__(self, report_id):
        super(ReportNotPermissioned, self).__init__(
            f"Not permissioned for report '{report_id}'"
        )


class ReportNotFound(EntityNotFound):
    def __init__(self, report_id):
        super(ReportNotFound, self).__init__(f"Report '{report_id}' not found")


class IOCNotFound(EntityNotFound):
    def __init__(self, report_id, ioc_id):
        super(IOCNotFound, self).__init__(f"IOC '{report_id}/{ioc_id}' not found")


class ReportEngine(object):
    VALID_ID_RE = re.compile(r"^[\w-]+$", re.ASCII)
    SOURCE_CLASSIFIER = "source"
    ACCESS_CLASSIFIER = "access"
    VIEW_CLASSIFIER = "view"

    PUBLIC_LABEL = "public"
    PRIVATE_LABEL = "private"
    RESERVED_LABEL = "reserved"

    def __init__(
        self, config: Config, db_pool: ReadWriteDbPool, metrics: MetricsRegistry
    ):
        self.config = config
        self.db_pool = db_pool
        self.metrics = metrics
        self.search_client = SearchClient(self.config)

    async def shutdown(self):
        await self.search_client.close()

    @staticmethod
    def gen_report_id() -> str:
        return (
            base64.b64encode(uuid.uuid4().bytes)
            .decode("utf-8")
            .rstrip("=\n")
            .replace("/", "")
            .replace("+", "")
        )

    @classmethod
    async def add_report(
        cls, conn, report: ApiReport, classifier_ids: List[int]
    ) -> str:
        db_report = cls._report_to_db(report)
        await conn.execute(DbReport.__table__.insert().values(**db_report.as_dict()))

        # Add IOCs
        if report.iocs_v2:
            for ioc in [
                cls._iocs_v2_to_db(ioc_v2, db_report.id) for ioc_v2 in report.iocs_v2
            ]:
                try:
                    await conn.execute(IOC.__table__.insert().values(**ioc.as_dict()))
                except IntegrityError:
                    raise ValueError(
                        "Duplicate IOC ID %s. IOC IDs must be unique within a report."
                        % ioc.id
                    )

        # add all classifiers for report
        for classifier_id in classifier_ids:
            await cls.insert_report_classifiers(conn, {db_report.id: [classifier_id]})

        return db_report.id

    async def get_reports_with_classifiers(
        self,
        classifier_ids: List[int],
        exclude_classifier_ids: List[int] = None,
        skip=0,
        take=0,
        get_count_only=False,
    ) -> Union[List[ApiReport], int]:
        sel_stmt = (
            select([ReportClassifier.report_id])
            .distinct(ReportClassifier.report_id)
            .where(ReportClassifier.classifier_id.in_(classifier_ids))
        )
        if exclude_classifier_ids:
            excluded_sel_stmt = (
                select([ReportClassifier.report_id])
                .distinct(ReportClassifier.report_id)
                .where(ReportClassifier.classifier_id.in_(exclude_classifier_ids))
            )
            sel_stmt = sel_stmt.except_(excluded_sel_stmt)

        sel_stmt = sel_stmt.order_by(ReportClassifier.report_id)
        if skip:
            sel_stmt = sel_stmt.offset(skip)

        if take:
            sel_stmt = sel_stmt.limit(take)

        if get_count_only:
            async with self.db_pool.ro_transaction() as conn:
                return (
                    await conn.execute(select(func.count()).select_from(sel_stmt))
                ).scalar()

        return await self.get_reports_from_ids(sel_stmt)

    async def get_classifier_id(
        self, classifier_key: str, classifier_value: str, conn=None
    ) -> int:
        if not conn:
            async with self.db_pool.ro_transaction() as conn:
                return await self._get_classifier_id(
                    conn, classifier_key, classifier_value
                )
        else:
            return await self._get_classifier_id(conn, classifier_key, classifier_value)

    @staticmethod
    async def _get_classifier_id(
        conn, classifier_key: str, classifier_value: str
    ) -> int:
        sel_stmt = (
            select([Classifier.id])
            .where(Classifier.key == classifier_key)
            .where(Classifier.value == classifier_value)
            .limit(1)
        )

        result = (await conn.execute(sel_stmt)).fetchone()

        if result is None:
            raise ValueError(f"Unknown classifier {classifier_key}/{classifier_value}")

        return result["id"]

    async def get_report_title(self, report_id: str, conn=None) -> str:
        if not conn:
            async with self.db_pool.ro_transaction() as conn:
                return await self._get_report_title(conn, report_id)
        else:
            return await self._get_report_title(conn, report_id)

    @staticmethod
    async def _get_report_title(conn, report_id: str) -> str:
        sel_stmt = select([DbReport.title]).where(DbReport.id == report_id).limit(1)

        result = (await conn.execute(sel_stmt)).fetchone()

        if result is None:
            raise ReportNotFound(report_id)

        return result["title"]

    async def get_reports_from_ids(
        self, ids_select: Union[str, Iterable[str]]
    ) -> List[ApiReport]:
        reports_dict = {}
        reports_dict_final = {}

        reports_subquery = (
            select([DbReport])
            .where(DbReport.id.in_(ids_select))
            .cte(f"{DbReport.__tablename__}")
        )
        reportAlias = aliased(DbReport, reports_subquery)

        sel_stmt = (
            select([reportAlias, IOC])
            .select_from(reports_subquery)
            .outerjoin(IOC.__table__, DbReport.id == IOC.report_id)
        )

        async with self.db_pool.ro_transaction() as conn:

            async for result in await conn.stream(sel_stmt):
                # result will contain duplicate keys for report, e.g. (report1.id, ioc1.id), (report1.id, ioc2.id), etc.
                #  pull out the first X items as the Report and the last items as the IOC
                report_result, ioc_result = result[0], result[1]

                if report_result is not None:
                    if reports_dict.get(report_result.id) is not None:
                        if ioc_result:
                            iocs = list(reports_dict[report_result.id].values())[0]
                            iocs.append(self.iocs_v2_to_api(ioc_result))
                            if reports_dict_final.get(report_result.id) is not None:
                                reports_dict_final[
                                    report_result.id
                                ] = self._report_to_api(report_result, iocs, len(iocs))
                    else:
                        temp_ioc = []
                        if ioc_result:
                            temp_ioc.append(self.iocs_v2_to_api(ioc_result))
                        reports_dict[report_result.id] = {report_result: temp_ioc}
                        reports_dict_final[report_result.id] = self._report_to_api(
                            report_result, temp_ioc, len(temp_ioc)
                        )

        return list(reports_dict_final.values())

    async def get_report(
        self, org_id: str, report_id: str, iocs_start_row=None, iocs_row_count=None
    ) -> ApiReport:
        async with self.db_pool.rw_transaction() as conn:
            return await self._get_report(
                conn, org_id, report_id, iocs_start_row, iocs_row_count
            )

    async def _get_report(
        self,
        conn,
        org_id: str,
        report_id: str,
        iocs_start_row=None,
        iocs_row_count=None,
    ) -> ApiReport:
        """
        Org must own the report being requested, on the view list, or be public

        classifiers
        ----
        source   |   ABC
        access   |   reserved
        view     |   DEF
        """
        classifier_ids = []
        try:
            classifier_ids.append(
                await self._get_classifier_id(conn, self.SOURCE_CLASSIFIER, org_id)
            )
        except ValueError:
            pass

        try:
            classifier_ids.append(
                await self._get_classifier_id(conn, self.VIEW_CLASSIFIER, org_id)
            )
        except ValueError:
            pass

        try:
            classifier_ids.append(
                await self._get_classifier_id(
                    conn, self.ACCESS_CLASSIFIER, self.PUBLIC_LABEL
                )
            )
        except ValueError:
            pass

        if not classifier_ids:
            raise ReportNotFound(report_id)

        join_stmt = DbReport.__table__.join(
            ReportClassifier.__table__, ReportClassifier.report_id == DbReport.id
        )

        sel_stmt = (
            select([DbReport])
            .select_from(join_stmt)
            .where(DbReport.id == report_id)
            .where(
                or_(
                    *[
                        ReportClassifier.classifier_id == classifier_id
                        for classifier_id in classifier_ids
                    ]
                )
            )
        )

        result = (await conn.execute(sel_stmt)).scalar()

        if not result:
            raise ReportNotFound(report_id)

        has_pagination_params = (
            iocs_start_row is not None and iocs_row_count is not None
        )

        # Get the IOCs on the report. Include pagination params if provided (+ total iocs count).
        iocs_v2 = []
        iocs_sel_stmt = select(IOC).where(IOC.report_id == report_id).order_by(IOC.id)
        if has_pagination_params:
            iocs_sel_stmt = iocs_sel_stmt.offset(iocs_start_row).limit(iocs_row_count)

        async for iocs_v2_result in await conn.stream(iocs_sel_stmt):
            iocs_v2.append(self.iocs_v2_to_api(iocs_v2_result[0]))

        # Lastly get the IOCs total count for the given report. Need for pagination handling in the UI.
        iocs_total_count = 0
        iocs_v2_length = len(iocs_v2)
        if not has_pagination_params or (
            iocs_start_row == 0 and iocs_v2_length < iocs_row_count
        ):
            # If there are no pagination params the total count is == len(iocs_v2)
            # If we are loading the first page and the results are less than iocs_row_count do not query the DB
            iocs_total_count = iocs_v2_length
        else:
            iocs_total_count = (
                await conn.execute(
                    select([func.count(IOC.id)]).where(IOC.report_id == report_id)
                )
            ).scalar()

        return self._report_to_api(result, iocs_v2, iocs_total_count)

    async def delete_report(self, conn, report_id: str) -> bool:
        return await self.delete_reports(conn, [report_id])

    async def delete_reports(
        self, conn, report_ids: Iterable[str], dirty_org_watchlists=True
    ) -> bool:
        # dirty the org watchlists for reports before we delete it
        #  otherwise we lose the watchlist associations on delete cascade
        if dirty_org_watchlists:
            await self.dirty_org_watchlist_for_reports(conn, report_ids)

        del_stmt = delete(DbReport).where(DbReport.id.in_(report_ids))

        return (await conn.execute(del_stmt)).rowcount > 0

    @staticmethod
    async def _update_report_timestamp(
        conn, report_id: str, report_timestamp: datetime.datetime
    ):
        update_stmt = (
            DbReport.__table__.update()
            .values(timestamp=report_timestamp)
            .where(DbReport.id == report_id)
            .where(DbReport.timestamp < report_timestamp)
        )

        result = await conn.execute(update_stmt)
        if result.rowcount == 0:
            raise ValueError("Timestamp older than existing report timestamp")

    async def delete_report_iocs(
        self,
        conn,
        report_id: str,
        ioc_ids: Iterable[str],
        report_timestamp: datetime.datetime,
    ) -> bool:
        await self._update_report_timestamp(conn, report_id, report_timestamp)

        del_stmt = (
            delete(IOC).where(IOC.report_id == report_id).where(IOC.id.in_(ioc_ids))
        )
        result = await conn.execute(del_stmt)
        return result.rowcount != 0

    async def add_report_iocs(
        self,
        conn,
        report_id: str,
        iocs: Iterable[IOC_V2],
        report_timestamp: datetime.datetime,
    ) -> bool:
        await self._update_report_timestamp(conn, report_id, report_timestamp)

        data = [
            {
                "id": ioc.id,
                "report_id": report_id,
                "match_type": ioc.match_type,
                "field": ioc.field,
                "values": json.dumps(ioc.values),
                "link": ioc.link,
            }
            for ioc in iocs
        ]

        insert_stmt = insert(IOC).values(data)

        try:
            result = await conn.execute(insert_stmt)
            return result.rowcount != 0
        except IntegrityError as e:
            if is_unique_violation_exception(e):
                _logger.error("Found duplicate key while adding report iocs_v2")
                raise ValueError("Unable to add report iocs_v2 due to id conflict")
            elif is_fk_violation_exception(e):
                _logger.error("Unable to add report iocs_v2")
                raise ValueError("Unable to add report iocs_v2 due to fk violation")
            else:
                raise e

    async def upsert_report(
        self, conn, report: ApiReport, classifier_ids: List[int] = None
    ) -> None:
        inserted_ids, updated_ids = await self._update_report_chunk(
            conn, [report], classifier_ids
        )
        if len(inserted_ids | updated_ids) == 0:
            raise ValueError(
                "Unable to update report because provided timestamp is out-of-date."
            )

        await self.dirty_org_watchlist_for_reports(
            conn, list(inserted_ids | updated_ids)
        )

    async def replace_reports(
        self, conn, reports: List[ApiReport], classifier_ids: List[int]
    ) -> Tuple[Set[str], Set[str]]:
        await self.metrics.histogram("feeds.reports.bulk_replace.size").add(
            len(reports)
        )

        with self.metrics.timer("reports.bulk_replace.duration").time():
            reports_chunks = chunker(1000, reports)
            all_inserted_ids, all_updated_ids = set(), set()

            for reports_chunk in reports_chunks:
                inserted_ids, updated_ids = await self._update_report_chunk(
                    conn, reports_chunk, classifier_ids
                )
                all_inserted_ids.update(inserted_ids)
                all_updated_ids.update(updated_ids)

            return all_inserted_ids, all_updated_ids

    async def dirty_org_watchlist_for_reports(
        self, conn, report_ids: Iterable[str], org_id: str = None
    ) -> None:
        if not report_ids:
            # nothing to do
            return

        reports_cte = (
            select(DbReport.id).where(DbReport.id.in_(report_ids)).cte("reports_cte")
        )

        # clear classifier watchlists
        join_stmt = ClassifierWatchlist.__table__.join(
            ReportClassifier.__table__,
            ReportClassifier.classifier_id == ClassifierWatchlist.classifier_id,
        )
        sel_stmt_1 = (
            select([ClassifierWatchlist.watchlist_id])
            .select_from(join_stmt)
            .where(ReportClassifier.report_id.in_(select([reports_cte.c.id])))
        )

        # clear report watchlists
        sel_stmt_2 = select([ReportWatchlist.watchlist_id]).where(
            ReportWatchlist.report_id.in_(select([reports_cte.c.id]))
        )

        await OrgWatchlistStateUpdater(
            self.config, get_utcnow_with_tzinfo()
        ).set_dirty_for_watchlist_id_select(
            sel_stmt_1.union(sel_stmt_2), org_id=org_id
        ).execute(
            conn
        )

    async def dirty_org_watchlist_for_classifier(
        self, conn, classifier_key: str, classifier_value: str
    ) -> None:
        report_join_stmt = ReportClassifier.__table__.join(
            Classifier.__table__, ReportClassifier.classifier_id == Classifier.id
        )
        report_sel_stmt = (
            select([ReportClassifier.report_id])
            .select_from(report_join_stmt)
            .where(Classifier.key == classifier_key)
            .where(Classifier.value == classifier_value)
        )

        # clear classifier watchlists
        join_stmt = ClassifierWatchlist.__table__.join(
            ReportClassifier.__table__,
            ReportClassifier.classifier_id == ClassifierWatchlist.classifier_id,
        )
        sel_stmt_1 = (
            select([ClassifierWatchlist.watchlist_id])
            .select_from(join_stmt)
            .where(ReportClassifier.report_id.in_(report_sel_stmt))
        )

        # clear report watchlists
        sel_stmt_2 = select([ReportWatchlist.watchlist_id]).where(
            ReportWatchlist.report_id.in_(report_sel_stmt)
        )

        await OrgWatchlistStateUpdater(
            self.config, get_utcnow_with_tzinfo()
        ).set_dirty_for_watchlist_id_select(sel_stmt_1.union(sel_stmt_2)).execute(conn)

    async def _update_report_chunk(
        self, conn, reports: Iterable[ApiReport], classifier_ids: Iterable[int]
    ) -> Tuple[Set[str], Set[str]]:
        reports_timestamps = {r.id: r.timestamp for r in reports}

        already_present_reports_stmt = select([DbReport.id, DbReport.timestamp]).where(
            DbReport.id.in_(reports_timestamps.keys())
        )
        already_present_reports = {
            r.id: r.timestamp.replace(tzinfo=datetime.timezone.utc)
            async for r in await conn.stream(already_present_reports_stmt)
        }

        missing_reports_ids = reports_timestamps.keys() - already_present_reports.keys()
        outdated_reports_ids = {
            rid
            for rid, rtimestamp in already_present_reports.items()
            if rtimestamp < reports_timestamps[rid]
        }

        report_ids_to_update = missing_reports_ids | outdated_reports_ids
        if not report_ids_to_update:
            return set(), set()

        db_reports_to_update = [
            self._report_to_db_dict(r) for r in reports if r.id in report_ids_to_update
        ]

        insert_stmt = insert(DbReport)
        upsert_stmt = insert_stmt.on_conflict_do_update(
            index_elements=[DbReport.id],
            set_={
                "timestamp": insert_stmt.excluded.timestamp,
                "link": insert_stmt.excluded.link,
                "title": insert_stmt.excluded.title,
                "severity": insert_stmt.excluded.severity,
                "tags": insert_stmt.excluded.tags,
                "description": insert_stmt.excluded.description,
                "visibility": insert_stmt.excluded.visibility,
            },
        )

        await conn.execute(upsert_stmt, db_reports_to_update)

        report_iocs_v2 = {
            r.id: r.iocs_v2
            for r in reports
            if r.iocs_v2 and r.id in report_ids_to_update
        }
        if report_iocs_v2:
            await self._remove_iocs_v2(conn, report_iocs_v2)
            await self._upsert_iocs_v2(conn, report_iocs_v2)

        if classifier_ids:
            # TODO: should we insert the classifiers for outdated_reports? Maybe we only need missing_reports here?
            report_classifiers = {
                report_id: classifier_ids for report_id in report_ids_to_update
            }
            if report_classifiers:
                await self.insert_report_classifiers(conn, report_classifiers)

        return missing_reports_ids, outdated_reports_ids

    async def _upsert_iocs_v2(
        self, conn, report_iocs_v2: Dict[str, List[IOC_V2]]
    ) -> None:
        data = [
            {
                "id": ioc_v2.id,
                "report_id": report_id,
                "match_type": ioc_v2.match_type,
                "field": ioc_v2.field,
                "values": json.dumps(ioc_v2.values),
                "link": ioc_v2.link,
            }
            for report_id, iocs_v2 in report_iocs_v2.items()
            for ioc_v2 in iocs_v2
        ]

        for data_chunk in chunker(1000, data):
            await self._upsert_iocs_v2_chunk(conn, data_chunk)

    @staticmethod
    async def _upsert_iocs_v2_chunk(conn, iocs_v2_data) -> None:
        insert_stmt = insert(IOC)
        upsert_stmt = insert_stmt.on_conflict_do_update(
            index_elements=[IOC.id, IOC.report_id],
            set_={
                "match_type": insert_stmt.excluded.match_type,
                "field": insert_stmt.excluded.field,
                "values": insert_stmt.excluded.get(
                    "values"
                ),  # need to do explicit column get here because
                #  'values' is a member of 'excluded'
                "link": insert_stmt.excluded.link,
            },
            # Only modify if there are any actual changes
            where=or_(
                IOC.match_type != insert_stmt.excluded.match_type,
                IOC.field != insert_stmt.excluded.field,
                IOC.values != insert_stmt.excluded.get("values"),
                IOC.link != insert_stmt.excluded.link,
            ),
        )

        try:
            await conn.execute(upsert_stmt, iocs_v2_data)
        except IntegrityError as e:
            if not is_fk_violation_exception(e):
                raise e

            _logger.error("Unable to update report iocs_v2")
            raise ValueError("Unable to update report iocs_v")

    @staticmethod
    async def _remove_iocs_v2(conn, report_iocs_v2: Dict[str, List[IOC_V2]]) -> None:
        sel_existing_iocs = select([IOC.id, IOC.report_id]).where(
            IOC.report_id.in_(report_iocs_v2.keys())
        )
        existing_iocs = {
            (ioc.id, ioc.report_id)
            async for ioc in await conn.stream(sel_existing_iocs)
        }

        new_iocs = {
            (ioc_v2.id, report_id)
            for report_id, iocs_v2 in report_iocs_v2.items()
            for ioc_v2 in iocs_v2
        }
        iocs_to_remove = existing_iocs - new_iocs

        for chunk in chunker(1000, iocs_to_remove):
            del_iocs_stmt = delete(IOC).where(tuple_(IOC.id, IOC.report_id).in_(chunk))
            await conn.execute(del_iocs_stmt)

    async def insert_classifiers(
        self, classifiers: List[Dict[str, str]], conn=None
    ) -> List[int]:
        if not conn:
            async with self.db_pool.rw_transaction() as conn:
                return await self._insert_classifiers(conn, classifiers)
        else:
            return await self._insert_classifiers(conn, classifiers)

    @staticmethod
    async def _insert_classifiers(conn, classifiers: List[Dict[str, str]]) -> List[int]:
        insert_stmt = insert(Classifier).values(classifiers).returning(Classifier.id)

        # Do update here so we get the ids back
        upsert_stmt = insert_stmt.on_conflict_do_update(
            index_elements=[Classifier.key, Classifier.value],
            set_={"key": insert_stmt.excluded.key},
        )

        ids = []
        for row in (await conn.execute(upsert_stmt)).fetchall():
            ids.append(row["id"])

        return ids

    async def verify_report_permissions(self, org_id: str, report_id: str) -> None:
        async with self.db_pool.rw_transaction() as conn:
            join_stmt = ReportClassifier.__table__.join(
                Classifier.__table__, Classifier.id == ReportClassifier.classifier_id
            )

            sel_stmt = (
                select([Classifier])
                .select_from(join_stmt)
                .where(
                    or_(
                        tuple_(Classifier.key, Classifier.value)
                        == (self.SOURCE_CLASSIFIER, org_id),
                        tuple_(Classifier.key, Classifier.value)
                        == (self.ACCESS_CLASSIFIER, self.PUBLIC_LABEL),
                    )
                )
                .where(ReportClassifier.report_id == report_id)
                .limit(2)
            )
            results = (await conn.execute(sel_stmt)).scalars()

            if not results:
                raise ReportNotFound(report_id)
            else:
                classifiers = {r.key: r.value for r in results}

                if classifiers.get(self.SOURCE_CLASSIFIER) != org_id:
                    if classifiers.get(self.ACCESS_CLASSIFIER) == self.PUBLIC_LABEL:
                        raise ReportNotPermissioned(report_id)
                    else:
                        raise ReportNotFound(report_id)

    async def validate_reports(self, org_id: str, reports: Iterable[ApiReport]):
        queries_to_validate = []
        for report in reports:
            queries_to_validate.extend(report.get_queries())
        if queries_to_validate:
            await self.validate_queries(org_id, queries_to_validate)

        fields_to_validate = []
        for report in reports:
            fields_to_validate.extend(report.get_equality_fields())
        if fields_to_validate:
            await self.validate_fields(org_id, fields_to_validate)

    async def validate_queries(self, org_id: str, queries: List[str]):
        first_invalid_query = await self.search_client.get_first_invalid_query(
            org_id, queries
        )

        if first_invalid_query is not None:
            raise InvalidQuery(first_invalid_query)

    async def validate_fields(self, org_id: str, iocs: List[IOC_V2]):
        for ioc in iocs:
            # validate ipv4 fields
            if ioc.field.endswith("ipv4"):
                for value in ioc.values:
                    try:
                        ipaddress.IPv4Address(value)
                    except ValueError:
                        # try IPv4 CIDR/network
                        try:
                            ipaddress.IPv4Network(value)
                        except ValueError:
                            raise InvalidField(ioc.field)

            # validate ip6 fields
            if ioc.field.endswith("ipv6"):
                for value in ioc.values:
                    try:
                        ipaddress.IPv6Address(value)
                    except ValueError:
                        raise InvalidField(ioc.field)

    @staticmethod
    def iocs_v2_to_api(record: IOC) -> IOC_V2:
        return IOC_V2(
            id=record.id,
            match_type=record.match_type,
            field=record.field,
            values=json.loads(record.values),
            link=record.link,
        )

    @classmethod
    def _report_to_api(
        cls,
        record: DbReport,
        iocs_v2: List[IOC_V2] = None,
        iocs_total_count: int = None,
        severity: int = None,
    ) -> ApiReport:
        return ApiReport(
            id=record.id,
            timestamp=record.timestamp.replace(tzinfo=datetime.timezone.utc),
            link=record.link,
            title=record.title,
            severity=severity if severity is not None else record.severity,
            tags=json.loads(record.tags) if record.tags else None,
            iocs_v2=iocs_v2 if iocs_v2 else None,
            iocs_total_count=iocs_total_count if iocs_total_count else 0,
            description=record.description,
            visibility=record.visibility,
        )

    @classmethod
    def _report_to_db(cls, record: ApiReport) -> DbReport:
        report = DbReport()
        report.id = record.id
        report.timestamp = record.timestamp
        report.link = record.link
        report.title = record.title
        report.severity = record.severity
        report.tags = json.dumps(record.tags) if record.tags else None
        report.description = record.description
        report.visibility = record.visibility

        return report

    @staticmethod
    def _report_to_db_dict(r: ApiReport):
        return {
            "id": r.id,
            "timestamp": r.timestamp,
            "link": r.link,
            "title": r.title,
            "severity": r.severity,
            "tags": json.dumps(r.tags) if r.tags else None,
            "description": r.description,
            "visibility": r.visibility,
        }

    @staticmethod
    def _iocs_v2_to_db(ioc_v2: IOC_V2, report_id: str) -> IOC:
        report_ioc = IOC()
        report_ioc.id = ioc_v2.id
        report_ioc.report_id = report_id
        report_ioc.match_type = ioc_v2.match_type
        report_ioc.field = ioc_v2.field
        report_ioc.values = json.dumps(ioc_v2.values)
        report_ioc.link = ioc_v2.link

        return report_ioc

    @staticmethod
    async def insert_report_classifiers(
        conn, report_classifiers: Dict[str, Iterable[int]]
    ):
        data = [
            {"report_id": report_id, "classifier_id": classifier_id}
            for report_id, classifier_ids in report_classifiers.items()
            for classifier_id in classifier_ids
        ]

        insert_stmt = insert(ReportClassifier)
        upsert_stmt = insert_stmt.on_conflict_do_nothing(
            index_elements=[ReportClassifier.report_id, ReportClassifier.classifier_id]
        )

        await conn.execute(upsert_stmt, data)

    @staticmethod
    async def remove_report_classifiers(
        conn, report_ids: Iterable[str], report_classifiers: Iterable[int]
    ):
        del_stmt = (
            delete(ReportClassifier)
            .where(ReportClassifier.report_id.in_(report_ids))
            .where(ReportClassifier.classifier_id.in_(report_classifiers))
        )

        await conn.execute(del_stmt)

    async def get_report_classifier_values(
        self, report_ids: Iterable[str], classifier_key: str
    ) -> Dict[str, str]:
        join_stmt = Classifier.__table__.join(
            ReportClassifier.__table__, ReportClassifier.classifier_id == Classifier.id
        )

        sel_stmt = (
            select([Classifier.value, ReportClassifier.report_id])
            .select_from(join_stmt)
            .where(Classifier.key == classifier_key)
            .where(ReportClassifier.report_id.in_(report_ids))
        )

        async with self.db_pool.rw_transaction() as conn:
            return {
                result["report_id"]: result["value"]
                async for result in await conn.stream(sel_stmt)
            }
