#!/usr/bin/env python3

import argparse
import logging
import sys
import asyncio
import os
import alembic.config

from feed_manager.sql.migrate import db_upgrade
from feed_manager.utils.config import Config

_logger = logging.getLogger(__name__)


async def initialize_db_schema(db_config):
    await db_upgrade(db_config)


def _get_config(args) -> Config:
    config = Config.create()

    config.set("db", "host", args.db_host)
    config.set("db", "port", str(args.db_port))
    config.set("db", "name", args.db_name)
    config.set("db", "user", args.db_user)
    config.set("db", "password", args.db_password)
    config.set("db", "ssl", args.db_ssl)

    return config


def build_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        required=False,
        default=False,
        help="Verbose logging",
    )
    parser.add_argument(
        "--db-host", dest="db_host", type=str, required=True, help="PostgreSQL host"
    )
    parser.add_argument(
        "--db-port",
        dest="db_port",
        type=int,
        required=False,
        default=5432,
        help="PostgreSQL port",
    )
    parser.add_argument(
        "--db-name",
        dest="db_name",
        type=str,
        required=True,
        help="PostgreSQL database name",
    )
    parser.add_argument(
        "--db-user", dest="db_user", type=str, required=True, help="PostgreSQL username"
    )
    parser.add_argument(
        "--db-password",
        dest="db_password",
        type=str,
        required=False,
        default=None,
        help="PostgreSQL password",
    )
    parser.add_argument(
        "--db-ssl",
        dest="db_ssl",
        type=str,
        required=False,
        default="verify-ca",
        help="PostgreSQL SSL mode",
    )
    parser.add_argument(
        "--init-db",
        action="store_true",
        dest="init_db",
        required=False,
        default=False,
        help="Initialize db",
    )
    parser.add_argument(
        "-r",
        "--revision",
        action="store_true",
        dest="revision",
        required=False,
        default=False,
        help="Create alembic revision from models",
    )
    parser.add_argument(
        "-m",
        "--message",
        type=str,
        dest="message",
        required=False,
        help="Revision message",
    )

    return parser


async def create_alembic_rev(config: Config, desc: str):  # pragma: no cover
    alembic_cfg = alembic.config.Config()
    alembic_cfg.set_main_option(
        "script_location",
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "../alembic"),
    )
    alembic_cfg.attributes["config"] = config
    alembic.config.command.revision(alembic_cfg, message=desc, autogenerate=True)


async def main():
    parser = build_parser()
    args = parser.parse_args()

    logging.basicConfig(
        stream=sys.stdout,
        level=logging.INFO,
        format="%(asctime)s.%(msecs)03d (%(levelname)s) - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    logging.getLogger().setLevel(logging.DEBUG if args.verbose else logging.INFO)

    config = _get_config(args)

    if args.init_db:
        await initialize_db_schema(config)
    elif args.revision:
        await initialize_db_schema(config)
        await create_alembic_rev(config, args.message)


if __name__ == "__main__":
    asyncio.run(main())
