from typing import List, Dict, Any, Iterable

import aiohttp
import logging

from feed_manager.api.models import WatchlistTelemetry

_logger = logging.getLogger(__name__)


class TelemetryClient(object):
    def __init__(self, config):
        self.config = config
        self.feed_search_url = self.config.get("feedsearch", "url")
        self.client_session = aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(
                total=self.config.getint("feedsearch", "timeout", fallback=60)
            )
        )

    async def close(self):
        await self.client_session.close()

    async def get_watchlist_telemetry(
        self, watchlist_id: str, intervals: Iterable[int]
    ) -> List[WatchlistTelemetry]:
        results = await self._post_telemetry(watchlist_id, intervals)

        return [
            WatchlistTelemetry(
                watchlist_id=watchlist_id,
                interval=result["interval"],
                hits=result["hits"],
                executions=result["executions"],
            )
            for result in results
        ]

    async def _post_telemetry(
        self, watchlist_id: str, intervals: Iterable[int]
    ) -> Dict[Any, Any]:
        payload = {"intervals": intervals}

        telemetry_url = (
            self.feed_search_url
            + f"/internal/feedsearch/v1/watchlist/{watchlist_id}/telemetry"
        )

        async with self.client_session.post(telemetry_url, json=payload) as resp:
            resp.raise_for_status()

            return await resp.json()
