import logging
import os
from typing import Collection

from aiohttp import web
import rapidjson as json
from aiohttp.web_exceptions import HTTPBadRequest

from feed_manager.api.models import FeedInfo, Report
from feed_manager.api.models.features import Features
from feed_manager.engine.feed_engine import FeedEngine
from feed_manager.engine.report_engine import ReportEngine

from feed_manager.utils import json_dumps

ORG_FIELD = os.environ.get("ORG_ID_FIELD", "org_key")

_logger = logging.getLogger(__name__)


def make_response_json(data=None) -> web.Response:
    if data is None:
        return web.Response(status=204)
    else:
        return web.json_response(data, dumps=json_dumps)


def get_request_org(request: web.Request):
    user_context = json.loads(request.headers["X-User-Context"])
    return str(user_context["target_org"][ORG_FIELD])


def validate_feed_scope(feed_scope):
    org_features = [feature.value for feature in Features]
    if feed_scope is not None and not all(
        scope in org_features for scope in feed_scope
    ):
        raise HTTPBadRequest(
            reason=f"Invalid feedinfo.scope. Org does not support specified scope"
        )


async def create_feed(
    engine: FeedEngine,
    org_key: str,
    params,
    required_access: str,
    delegates: Collection[str] = None,
) -> web.Response:
    if "feedinfo" not in params:
        raise HTTPBadRequest(reason="Missing feedinfo")

    if "reports" not in params:
        raise HTTPBadRequest(reason="Missing reports")

    params["feedinfo"]["owner"] = org_key

    req_access = params["feedinfo"].get("access", None)
    if req_access and req_access != required_access:
        raise HTTPBadRequest(
            reason=f"feedinfo.access must be '{required_access}' for this request"
        )

    feed_scope = params["feedinfo"].get("scope", None)

    if required_access != ReportEngine.PUBLIC_LABEL and feed_scope is not None:
        raise HTTPBadRequest(
            reason=f"feedinfo.scope is not applicable for private or reserved feed"
        )

    if feed_scope is not None:
        validate_feed_scope(feed_scope)

    params["feedinfo"]["access"] = required_access

    feedinfo = await engine.create_feed(
        org_key,
        FeedInfo.from_dict(params["feedinfo"]),
        [Report.from_dict(report) for report in params["reports"]],
        delegates=delegates,
    )

    return make_response_json(FeedEngine.remove_empty_feed_scope(feedinfo))
